;;;----------------------------------------------------------------------
;;; Eleven compiler php back end (from intermediate code file to
;;; finished PHP program).
;;; 
;;; Copyright (C) 2004 Joseph Morrison <jdm@statesafe.org>.
;;; 
;;; This program is free software; you can redistribute it and/or modify it under
;;; the terms of the GNU General Public License as published by the Free Software
;;; Foundation; either version 2 of the License, or (at your option) any later
;;; version.
;;; 
;;; This program is distributed in the hope that it will be useful, but WITHOUT
;;; ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;;; details. Visit http://www.gnu.org/copyleft/gpl.html for the full text of the
;;; license, or you can request a copy from the Free Software Foundation, Inc.,
;;; 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
;;; 
;;; This license allows you to use the Eleven compiler no charge under the
;;; condition that if you create a derived work (either based on the Eleven
;;; compiler or based on code generated by the Eleven compiler) and redistribute
;;; it, the complete source code for your application must be available and freely
;;; redistributable.  A commercial license for Eleven is available without this
;;; condition. For more information please email info@statesafe.org.
;;;----------------------------------------------------------------------

(define (main)

  ;; Get command line arguments (name of this program, data directory,
  ;; and input filename).

  (if (not (= (length (command-line)) 3))
      (begin
	(display "c11: internal error starting compiler third stage")
	(newline)
	(exit 1)))

  (let ((program-self   (list-ref (command-line) 0))
	(data-directory (list-ref (command-line) 1))
	(input-file     (list-ref (command-line) 2)))

    ;; Read intermediate code from file (compilation parameters,
    ;; schema information, and program instructions).

    (let ((intermediate-code (with-input-from-file input-file read)))

      (let ((app-name       (get-value 'app-name       intermediate-code))
	    (package-name   (get-value 'package-name   intermediate-code))
	    (auth-method    (get-value 'auth-method    intermediate-code))
	    (db-connect     (get-value 'db-connect     intermediate-code))
	    (db-user 	    (get-value 'db-user        intermediate-code))
	    (db-pass 	    (get-value 'db-pass        intermediate-code))
	    (style-sheet    (get-value 'style-sheet    intermediate-code))
	    (fav-icon 	    (get-value 'fav-icon       intermediate-code))
	    (display-header (get-value 'display-header intermediate-code))
	    (schema-id 	    (get-value 'schema-id      intermediate-code))
	    (schema-entries (get-value 'schema-entries intermediate-code))
	    (transient-vars (get-value 'transient-vars intermediate-code))
	    (instructions   (get-value 'instructions   intermediate-code)))

	;; Translate into finished PHP program.

	(compile-write-final data-directory
			     app-name
			     package-name
			     auth-method
			     db-connect
			     db-user
			     db-pass
			     style-sheet
			     fav-icon
			     display-header
			     schema-id
			     schema-entries
			     transient-vars
			     instructions)))))

;;;----------------------------------------------------------------------
;;; Read the requested value from the intermediate code.
;;;----------------------------------------------------------------------

(define (get-value key alist)
  (let ((value (assq key alist)))
    (if value
	(cadr value)
	(begin
	  (display "c11: missing \"")
	  (display key)
	  (display "\" value in compiler third stage")
	  (newline)
	  (exit 1)))))
	
;;;----------------------------------------------------------------------
;;; Translate instructions to PHP and write output file.
;;;----------------------------------------------------------------------

(define (compile-write-final data-directory
			     app-name
			     package-name
			     auth-method
			     db-connect
			     db-user
			     db-pass
			     style-sheet
			     fav-icon
			     display-header
			     schema-id
			     schema-entries
			     transient-vars
			     instructions)

  (with-output-to-file (string-append package-name ".php")
    (lambda ()
      (display "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">")
      (newline)
      (display "<?php")
      (newline)
      (newline)
      (display "#----------------------------------------------------------------------")
      (newline)
      (display "# Program \"")
      (display app-name)
      (display "\" (package \"")
      (display package-name)
      (display "\").")
      (newline)
      (display "# ")
      (newline)
      (display "# Generated by the Eleven compiler on ")
      (display (strftime "%c" (localtime (current-time))))
      (display ".")
      (newline)
      (display "# Runtime requires Apache, PHP, and MySQL.")
      (newline)
      (display "# ")
      (newline)
      (display "# The Eleven compiler is copyright (2004) Joseph Morrison <jdm@statesafe.org>")
      (newline)
      (display "# and is distributed under the terms of the General Public License (GPL).")
      (newline)
      (display "# Visit http://www.gnu.org/copyleft/gpl.html for details.")
      (newline)
      (display "# ")
      (newline)
      (display "# This compiled program is considered a derivative work of the Eleven compiler,")
      (newline)
      (display "# therefore it is also covered by the GPL. Briefly, that means that you may use")
      (newline)
      (display "# it freely (including for commercial use) but if you redistribute it, you must")
      (newline)
      (display "# make source code available. For information about unrestricted commercial")
      (newline)
      (display "# licenses, please email info@statesafe.org.")
      (newline)
      (display "#----------------------------------------------------------------------")
      (newline)
      (newline)
      (display "?>")
      (newline)
      (newline)

      (with-input-from-file (string-append data-directory "/c11-prefix.php")
        (lambda ()
          (let loop ((next-character (read-char)))
            (if (not (eof-object? next-character))
                (begin
                  (write-char next-character)
                  (loop (read-char)))))))

      (newline)
      (display "# If there are any \"include\" statements in the source program, they are listed here:")
      (newline)

      (for-each (lambda (i)
		  (if (and (pair? i) (eq? (car i) 'include))
		      (let ((include-item (cadr i)))
			(display include-item)
			(newline))))
		instructions)

      (newline)
      (display "function generate_display_missing_state ($status_message)")
      (newline)
      (display "{")
      (newline)
      (write-generate-display-code-builtin app-name auth-method style-sheet fav-icon display-header '())
      (display "}")
      (newline)
      (newline)
      (display "function generate_display ($status_message, &$state)")
      (newline)
      (display "{")
      (newline)
      (write-generate-display-code app-name package-name auth-method style-sheet fav-icon display-header schema-id instructions)
      (display "}")
      (newline)
      (newline)
      (display "function execute_program ($dbh, &$parameters, &$state)")
      (newline)
      (display "{")
      (newline)
      (display "  $next_state = $state;")
      (newline)
      (display "  $next_state ['resumed_session'] = 'no';")
      (newline)
      (newline)
      (display "  $target = \"resume_from_display_\" . $next_state ['display_state'];")
      (newline)
      (newline)
      (display "  while (TRUE)")
      (newline)
      (display "  {")
      (newline)
      (display "   switch ($target)")
      (newline)
      (display "   {")
      (newline)
      (write-program-code instructions #f package-name schema-id)
      (display "   }")
      (newline)
      (display "  }")
      (newline)
      (display "}")
      (newline)
      (newline)
      (write-schema-update-code schema-id schema-entries)
      (newline)
      (display "function handler ()")
      (newline)
      (display "{")
      (newline)
      (display "  runtime11_handler (\"")
      (display db-connect)
      (display "\", \"")
      (display db-user)
      (display "\", \"")
      (display db-pass)
      (display "\", \"")
      (display package-name)
      (display "\", \"")
      (display auth-method)
      (display "\", \"")
      (display schema-id)
      (display "\");")
      (newline)
      (display "}")
      (newline)
      (newline)

      (with-input-from-file (string-append data-directory "/c11-suffix.php")
        (lambda ()
          (let loop ((next-character (read-char)))
            (if (not (eof-object? next-character))
                (begin
                  (write-char next-character)
                  (loop (read-char)))))))))

  (exit 0))

;;;----------------------------------------------------------------------
;;; Emit the PHP code that generates all of the XHTML displays
;;; (to be inserted in the generate_display function).
;;;----------------------------------------------------------------------

(define (write-generate-display-code app-name package-name auth-method style-sheet fav-icon display-header schema-id instructions)

  (for-each (lambda (i)
	      (if (and (pair? i) (eq? (car i) 'display))

		  (let ((display-id (list-ref i 1))
			(display-name (list-ref i 2))
			(display-option (list-ref i 3))
			(display-instructions (list-tail i 4)))

		    (newline)

		    ;; if the display id is 0, 1, or 2, generate the corresponding predefined display,
		    ;; otherwise generate the user-specified display from the program code

		    (if (or (= display-id 0) (= display-id 1) (= display-id 2))

			(write-generate-display-code-builtin app-name
							     auth-method
							     style-sheet
							     fav-icon
							     display-header
							     display-id)

			(write-generate-display-code-general app-name
							     package-name
							     auth-method
							     style-sheet
							     fav-icon
							     display-header
							     schema-id
							     display-instructions
							     display-id
							     display-name
							     display-option)))))
	    instructions))

;;;----------------------------------------------------------------------
;;; Emit the PHP code for a predefined display.
;;;----------------------------------------------------------------------

(define (write-generate-display-code-builtin app-name
					     auth-method
					     style-sheet
					     fav-icon
					     display-header
					     display-id)

  (cond ((null? display-id)
	 (display "  # Generate error display if missing state hash:")
	 (newline)
	 (newline)
	 (display "  if (TRUE)")
	 (newline))

	((= display-id 0)
	 (display "  # If display state is 0, generate login display:")
	 (newline)
	 (newline)
	 (display "  if ($state ['display_state'] == ")
	 (display display-id)
	 (display ")")
	 (newline))

	((= display-id 1)
	 (display "  # If display state is 1, generate bad login display:")
	 (newline)
	 (newline)
	 (display "  if ($state ['display_state'] == ")
	 (display display-id)
	 (display ")")
	 (newline))

	((= display-id 2)
	 (display "  # If display state is 2, generate end session display:")
	 (newline)
	 (newline)
	 (display "  if ($state ['display_state'] == ")
	 (display display-id)
	 (display ")")
	 (newline)))

  (display "  {")
  (newline)
  (display "    echo \"  <head>\\n\";")
  (newline)
  (display "    echo \"    <title>")
  (display app-name)
  (display "</title>\\n\";")
  (newline)
  (display "    echo \"    <meta http-equiv=\\\"Content-Type\\\" content=\\\"text/html; charset=ISO-8859-1\\\" />\\n\";")
  (newline)
  (display "    echo \"    <meta name=\\\"generator\\\" content=\\\"Eleven (see http://eleven.sourceforge.net for more information)\\\" />\\n\";")
  (newline)
  (display "    echo \"    <link rel=\\\"stylesheet\\\" type=\\\"text/css\\\" href=\\\"")
  (display style-sheet)
  (display "\\\" />\\n\";")
  (newline)
  (display "    echo \"    <link rel=\\\"shortcut icon\\\" type=\\\"image/x-icon\\\" href=\\\"")
  (display fav-icon)
  (display "\\\" />\\n\";")
  (newline)
  (display "    echo \"  </head>\\n\";")
  (newline)
  (display "    echo \"  <body>\\n\";")
  (newline)

  (if (not (null? display-id))
      (begin
	(display "    echo \"    <form action=\\\"\" . $_SERVER['PHP_SELF'] . \"?cid=\" . $state ['cid'] . \"\\\" method=\\\"post\\\">\\n\";")
	(newline)
	(display "    echo \"      <input type=\\\"hidden\\\" name=\\\"authtoken\\\" value=\\\"\" . $state ['auth_token'] . \"\\\" />\\n\";")
	(newline)
	(display "    echo \"      <input type=\\\"hidden\\\" name=\\\"control\\\" value=\\\"\\\" />\\n\";")
	(newline)))

  (display "    echo \"      <table summary=\\\"outermost table (single column, rows correspond to header, main content, and footer)\\\" width=\\\"100%\\\">\\n\";")
  (newline)

  (if (not (string=? display-header "no"))
      (begin
	(display "    echo \"        <tr>\\n\";")
	(newline)
	(display "    echo \"          <td class=\\\"pageAutoHeader\\\">\\n\";")
	(newline)
	(display "    echo \"            <p class=\\\"statusMessage\\\">&nbsp;$status_message</p>\\n\";")
	(newline)
	(display "    echo \"            <hr />\\n\";")
	(newline)
	(display "    echo \"          </td>\\n\";")
	(newline)
	(display "    echo \"        </tr>\\n\";")
	(newline)))

  (display "    echo \"        <tr>\\n\";")
  (newline)
  (display "    echo \"          <td class=\\\"pageMain\\\">\\n\";")
  (newline)

  (cond ((null? display-id)
	 (display "    echo \"            <h1>")
	 (display app-name)
	 (display " - internal error</h1>\\n\";")
	 (newline)
	 (display "    echo \"            <p>Internal error - missing state reference.</p>\\n\";")
	 (newline))

	((= display-id 0)
	 (display "    echo \"            <h1>")
	 (display app-name)
	 (display "</h1>\\n\";")
	 (newline)
	 (display "    echo \"            <table summary=\\\"login display (login and password entry fields)\\\">\\n\";")
	 (newline)
	 (display "    echo \"              <tr>\\n\";")
	 (newline)
	 (display "    echo \"                <td>Login:</td>\\n\";")
	 (newline)
	 (display "    echo \"                <td><input type=\\\"text\\\" name=\\\"login\\\" value=\\\"\\\" /></td>\\n\";")
	 (newline)
	 (display "    echo \"              </tr>\\n\";")
	 (newline)
	 (display "    echo \"              <tr>\\n\";")
	 (newline)
	 (display "    echo \"                <td>Password:</td>\\n\";")
	 (newline)
	 (display "    echo \"                <td><input type=\\\"password\\\" name=\\\"password\\\" value=\\\"\\\" /></td>\\n\";")
	 (newline)
	 (display "    echo \"              </tr>\\n\";")
	 (newline)
	 (display "    echo \"            </table>\\n\";")
	 (newline)
	 (display "    echo \"            <p>\\n\";")
	 (newline)
	 (display "    echo \"              <input class=\\\"controlButton\\\" type=\\\"submit\\\" name=\\\"action\\\" value=\\\"Continue\\\" />\\n\";")
	 (newline)
	 (display "    echo \"            </p>\\n\";")
	 (newline))

	((= display-id 1)
	 (display "    echo \"            <h1>")
	 (display app-name)
	 (display "</h1>\\n\";")
	 (newline)
	 (display "    echo \"            <p>Account does not exist or password is incorrect.</p>\\n\";")
	 (newline)
	 (display "    echo \"            <p>\\n\";")
	 (newline)
	 (display "    echo \"              <input class=\\\"controlButton\\\" type=\\\"submit\\\" name=\\\"action\\\" value=\\\"Continue\\\" />\\n\";")
	 (newline)
	 (display "    echo \"            </p>\\n\";")
	 (newline))

	((= display-id 2)
	 (display "    echo \"            <p>End of session.</p>\\n\";")
	 (newline)
	 (cond ((string=? auth-method "auto")
		(begin
		  (display "    echo \"            <p>\\n\";")
		  (newline)
		  (display "    echo \"              <input class=\\\"controlButton\\\" type=\\\"submit\\\" name=\\\"action\\\" value=\\\"Start new session\\\" />\\n\";")
		  (newline)
		  (display "    echo \"            </p>\\n\";")
		  (newline)))
	       (else
		(begin
		  (display "    echo \"            <p>\\n\";")
		  (newline)
		  (display "    echo \"              <input class=\\\"controlButton\\\" type=\\\"submit\\\" name=\\\"action\\\" value=\\\"Start new session\\\" />\\n\";")
		  (newline)
		  (display "    echo \"              <input class=\\\"controlButton\\\" type=\\\"submit\\\" name=\\\"action\\\" value=\\\"Logout\\\" />\\n\";")
		  (newline)
		  (display "    echo \"            </p>\\n\";")
		  (newline))))))

  (display "    echo \"          </td>\\n\";")
  (newline)
  (display "    echo \"        </tr>\\n\";")
  (newline)
  (display "    echo \"        <tr>\\n\";")
  (newline)
  (display "    echo \"          <td class=\\\"pageAutoFooter\\\">\\n\";")
  (newline)
  (display "    echo \"            <hr />\\n\";")
  (newline)
  (display "    echo \"            <p class=\\\"poweredByMessage\\\">\\n\";")
  (newline)
  (display "    echo \"              <a href=\\\"http://eleven.sourceforge.net/\\\"><span class=\\\"poweredByLeft\\\">&nbsp;Powered&nbsp;by&nbsp;</span><span class=\\\"poweredByRight\\\">Eleven</span></a>\\n\";")
  (newline)
  (display "    echo \"            </p>\\n\";")
  (newline)
  (display "    echo \"          </td>\\n\";")
  (newline)
  (display "    echo \"        </tr>\\n\";")
  (newline)
  (display "    echo \"      </table>\\n\";")
  (newline)

  (if (not (null? display-id))
      (begin
	(display "    echo \"    </form>\\n\";")
	(newline)))

  (display "    echo \"  </body>\\n\";")
  (newline)
  (display "  }")
  (newline))

;;;----------------------------------------------------------------------
;;; Emit the PHP code for a user-specified display.
;;;----------------------------------------------------------------------

(define (write-generate-display-code-general app-name
					     package-name
					     auth-method
					     style-sheet
					     fav-icon
					     display-header
					     schema-id
					     display-instructions
					     display-id
					     display-name
					     display-option)
  (if (string=? display-name "anonymous")
      (begin
	(display "  # If display state is ")
	(display display-id)
	(display ", generate the following unnamed display:")
	(newline))
      (begin
	(display "  # If display state is ")
	(display display-id)
	(display ", generate \"")
	(display display-name)
	(display "\" display:")
	(newline)))
  (newline)
  (display "  if ($state ['display_state'] == ")
  (display display-id)
  (display ")")
  (newline)
  (display "  {")
  (newline)
  (display "   $target = \"display_start\";")
  (newline)
  (display "   $not_finished = TRUE;")
  (newline)
  (display "   while ($not_finished)")
  (newline)
  (display "   {")
  (newline)
  (display "    switch ($target)")
  (newline)
  (display "    {")
  (newline)
  (display "    case \"display_start\":")
  (newline)
  (display "    echo \"  <head>\\n\";")
  (newline)
  (display "    echo \"    <title>")
  (display app-name)
  (if (not (string=? display-name "anonymous"))
      (begin
	(display " (")
	(display display-name)
	(display ")")))
  (display "</title>\\n\";")
  (newline)
  (display "    echo \"    <meta http-equiv=\\\"Content-Type\\\" content=\\\"text/html; charset=ISO-8859-1\\\" />\\n\";")
  (newline)
  (display "    echo \"    <meta name=\\\"generator\\\" content=\\\"Eleven (see http://eleven.sourceforge.net for more information)\\\" />\\n\";")
  (newline)
  (display "    echo \"    <link rel=\\\"stylesheet\\\" type=\\\"text/css\\\" href=\\\"")
  (display style-sheet)
  (display "\\\" />\\n\";")
  (newline)
  (display "    echo \"    <link rel=\\\"shortcut icon\\\" type=\\\"image/x-icon\\\" href=\\\"")
  (display fav-icon)
  (display "\\\" />\\n\";")
  (newline)
  (display "    echo \"  </head>\\n\";")
  (newline)
  (newline)
  (display "    $reference_html = \"                  <table summary=\\\"two-column table per display area (labels on left, entry fields on right)\\\" width=\\\"100%\\\">\\n\";")
  (newline)
  (display "    $reference_html   .= \"                    <colgroup>\\n\";")
  (newline)
  (display "    $reference_html   .= \"                      <col width=\\\"15%\\\" />\\n\";")
  (newline)
  (display "    $reference_html   .= \"                      <col width=\\\"85%\\\" />\\n\";")
  (newline)
  (display "    $reference_html   .= \"                    </colgroup>\\n\";")
  (newline)
  (newline)

  (display "    $header_html = $reference_html;")
  (newline)
  (display "    $left_html   = $reference_html;")
  (newline)
  (display "    $main_html   = $reference_html;")
  (newline)
  (display "    $right_html  = $reference_html;")
  (newline)
  (display "    $footer_html = $reference_html;")
  (newline)
  (newline)
  
  (write-program-code display-instructions #t package-name schema-id)
  (newline)

  (display "    $postscript = \"                  </table>\\n\";")
  (newline)
  (newline)

  (display "    $header_html .= $postscript;")
  (newline)
  (display "    $left_html   .= $postscript;")
  (newline)
  (display "    $main_html   .= $postscript;")
  (newline)
  (display "    $right_html  .= $postscript;")
  (newline)
  (display "    $footer_html .= $postscript;")
  (newline)
  (display "    $reference_html .= $postscript;")
  (newline)
  (newline)

  (display "    echo \"  <body>\\n\";")
  (newline)
  (display "    echo \"    <form action=\\\"\" . $_SERVER['PHP_SELF'] . \"?cid=\" . $state ['cid'] . \"\\\" method=\\\"post\\\">\\n\";")
  (newline)
  (display "    echo \"      <input type=\\\"hidden\\\" name=\\\"authtoken\\\" value=\\\"\" . $state ['auth_token'] . \"\\\" />\\n\";")
  (newline)
  (display "    echo \"      <input type=\\\"hidden\\\" name=\\\"control\\\" value=\\\"\\\" />\\n\";")
  (newline)
  (display "    echo \"      <table summary=\\\"outermost table (single column, rows correspond to header, main content, and footer)\\\" width=\\\"100%\\\">\\n\";")
  (newline)

  (if (not (string=? display-header "no"))
      (begin
	(display "    echo \"        <tr>\\n\";")
	(newline)
	(display "    echo \"          <td class=\\\"pageAutoHeader\\\">\\n\";")
	(newline)
	(display "    echo \"            <p class=\\\"statusMessage\\\">&nbsp;$status_message</p>\\n\";")
	(newline)
	(display "    echo \"            <hr />\\n\";")
	(newline)
	(display "    echo \"          </td>\\n\";")
	(newline)
	(display "    echo \"        </tr>\\n\";")
	(newline)))

  (display "    echo \"        <tr>\\n\";")
  (newline)
  (display "    echo \"          <td class=\\\"pageUserHeader\\\">\\n\";")
  (newline)
  (display "    echo $header_html;")
  (newline)
  (display "    echo \"          </td>\\n\";")
  (newline)
  (display "    echo \"        </tr>\\n\";")
  (newline)
  (display "    echo \"        <tr>\\n\";")
  (newline)
  (display "    echo \"          <td class=\\\"pageMiddle\\\">\\n\";")
  (newline)
  (display "    echo \"            <table summary=\\\"three-column table for page middle section (left sidebar, main area, right sidebar)\\\" width=\\\"100%\\\">\\n\";")
  (newline)
  (display "    echo \"              <colgroup>\\n\";")
  (newline)
  (display "    echo \"                <col width=\\\"15%\\\" />\\n\";")
  (newline)
  (display "    echo \"                <col width=\\\"70%\\\" />\\n\";")
  (newline)
  (display "    echo \"                <col width=\\\"15%\\\" />\\n\";")
  (newline)
  (display "    echo \"              </colgroup>\\n\";")
  (newline)
  (display "    echo \"              <tr>\\n\";")
  (newline)
  (display "    if ($left_html === $reference_html)")
  (newline)
  (display "    {")
  (newline)
  (display "      echo \"                <td class=\\\"pageMain\\\" colspan=\\\"2\\\">\\n\";")
  (newline)
  (display "      echo $main_html;")
  (newline)

  (cond ((string=? auth-method "auto")
	 (if (not (eq? display-option 'no_continue))
	     (begin
	       (display "      echo \"                  <p>\\n\";")
	       (newline)
	       (display "      echo \"                    <input class=\\\"controlButton\\\" type=\\\"submit\\\" name=\\\"action\\\" value=\\\"Continue\\\" />\\n\";")
	       (newline)
	       (display "      echo \"                  </p>\\n\";")
	       (newline))))
 
	(else
	 (begin
	   (display "      echo \"                  <p>\\n\";")
	   (newline)
	   (if (not (eq? display-option 'no_continue))
	       (begin
		 (display "      echo \"                    <input class=\\\"controlButton\\\" type=\\\"submit\\\" name=\\\"action\\\" value=\\\"Continue\\\" />\\n\";")
		 (newline)))
	   (display "      echo \"                    <input class=\\\"controlButton\\\" type=\\\"submit\\\" name=\\\"action\\\" value=\\\"Logout\\\" />\\n\";")
	   (newline)
	   (display "      echo \"                  </p>\\n\";")
	   (newline))))

  (display "      echo \"                </td>\\n\";")
  (newline)
  (display "      echo \"                <td class=\\\"pageRight\\\">\\n\";")
  (newline)
  (display "      echo $right_html;")
  (newline)
  (display "      echo \"                </td>\\n\";")
  (newline)
  (display "    }")
  (newline)
  (display "    else")
  (newline)
  (display "    {")
  (newline)
  (display "      echo \"                <td class=\\\"pageLeft\\\">\\n\";")
  (newline)
  (display "      echo $left_html;")
  (newline)

  (cond ((string=? auth-method "auto") #f)
	(else (begin
		(display "      echo \"                  <p>\\n\";")
		(newline)
		(display "      echo \"                    <input class=\\\"controlButton\\\" type=\\\"submit\\\" name=\\\"action\\\" value=\\\"Logout\\\" />\\n\";")
		(newline)
		(display "      echo \"                  </p>\\n\";")
		(newline))))

  (display "      echo \"                </td>\\n\";")
  (newline)
  (display "      echo \"                <td class=\\\"pageMain\\\">\\n\";")
  (newline)
  (display "      echo $main_html;")
  (newline)

  (if (not (eq? display-option 'no_continue))
      (begin
	(display "      echo \"                  <p>\\n\";")
	(newline)
	(display "      echo \"                    <input class=\\\"controlButton\\\" type=\\\"submit\\\" name=\\\"action\\\" value=\\\"Continue\\\" />\\n\";")
	(newline)
	(display "      echo \"                  </p>\\n\";")
	(newline)))

  (display "      echo \"                </td>\\n\";")
  (newline)
  (display "      echo \"                <td class=\\\"pageRight\\\">\\n\";")
  (newline)
  (display "      echo $right_html;")
  (newline)
  (display "      echo \"                </td>\\n\";")
  (newline)
  (display "    }")
  (newline)
  (display "    echo \"              </tr>\\n\";")
  (newline)
  (display "    echo \"            </table>\\n\";")
  (newline)
  (display "    echo \"          </td>\\n\";")
  (newline)
  (display "    echo \"        </tr>\\n\";")
  (newline)
  (display "    echo \"        <tr>\\n\";")
  (newline)
  (display "    echo \"          <td class=\\\"pageUserFooter\\\">\\n\";")
  (newline)
  (display "    echo $footer_html;")
  (newline)
  (display "    echo \"          </td>\\n\";")
  (newline)
  (display "    echo \"        </tr>\\n\";")
  (newline)
  (display "    echo \"        <tr>\\n\";")
  (newline)
  (display "    echo \"          <td class=\\\"pageAutoFooter\\\">\\n\";")
  (newline)
  (display "    echo \"            <hr />\\n\";")
  (newline)
  (display "    echo \"            <p class=\\\"poweredByMessage\\\">\\n\";")
  (newline)
  (display "    echo \"              <a href=\\\"http://eleven.sourceforge.net/\\\"><span class=\\\"poweredByLeft\\\">&nbsp;Powered&nbsp;by&nbsp;</span><span class=\\\"poweredByRight\\\">Eleven</span></a>\\n\";")
  (newline)
  (display "    echo \"            </p>\\n\";")
  (newline)
  (display "    echo \"          </td>\\n\";")
  (newline)
  (display "    echo \"        </tr>\\n\";")
  (newline)
  (display "    echo \"      </table>\\n\";")
  (newline)
  (display "    echo \"    </form>\\n\";")
  (newline)
  (display "    echo \"  </body>\\n\";")
  (newline)
  (display "    $not_finished = FALSE;")
  (newline)
  (display "    }")
  (newline)
  (display "   }")
  (newline)
  (display "  }")
  (newline))

;;;----------------------------------------------------------------------
;;; Emit the PHP code that performs all of the computation between
;;; XHTML displays (to be inserted in the execute_program function).
;;;----------------------------------------------------------------------

(define (write-program-code instructions in-display? package-name schema-id)
  (for-each (lambda (i)
	      (write-inst i in-display? package-name schema-id))
	    instructions))

;;;----------------------------------------------------------------------
;;; Emit the PHP code for a single instruction.
;;;----------------------------------------------------------------------

(define (write-inst i in-display? package-name schema-id)

  ;; Instructions that can be used in any context:

  (cond ((symbol? i) 		      (write-inst-label i))
	((string? i) 		      (write-inst-label i))

	((eq? (car i) 'bogus)         	    (write-inst-bogus        	(cdr i)))
	((eq? (car i) 'comment)        	    (write-inst-comment        	(cdr i)))
	((eq? (car i) 'include)        	    (write-inst-include        	(cdr i)))
	((eq? (car i) 'declare-temp)  	    (write-inst-declare-temp 	(cdr i)))
	((eq? (car i) 'assign)        	    (write-inst-assign       	(cdr i)))
	((eq? (car i) 'append)        	    (write-inst-append       	(cdr i)))
	((eq? (car i) 'increment)     	    (write-inst-increment    	(cdr i)))
	((eq? (car i) 'decrement)     	    (write-inst-decrement    	(cdr i)))
	((eq? (car i) 'multiply)      	    (write-inst-multiply     	(cdr i)))
	((eq? (car i) 'divide)        	    (write-inst-divide       	(cdr i)))
	((eq? (car i) 'substring)     	    (write-inst-substring    	(cdr i)))
	((eq? (car i) 'number-equal)  	    (write-inst-number-equal 	(cdr i)))
	((eq? (car i) 'number-less)    	    (write-inst-number-less 	(cdr i)))
	((eq? (car i) 'number-greater) 	    (write-inst-number-greater  (cdr i)))
	((eq? (car i) 'string-equal)  	    (write-inst-string-equal 	(cdr i)))
	((eq? (car i) 'string-match)  	    (write-inst-string-match 	(cdr i)))
	((eq? (car i) 'and)                 (write-inst-and             (cdr i)))
	((eq? (car i) 'or)                  (write-inst-or              (cdr i)))
	((eq? (car i) 'not)                 (write-inst-not             (cdr i)))
	((eq? (car i) 'sqrt)                (write-inst-sqrt            (cdr i)))
	((eq? (car i) 'goto)          	    (write-inst-goto         	(cdr i)))
	((eq? (car i) 'goto-if-true)  	    (write-inst-goto-if-true 	(cdr i)))
	((eq? (car i) 'goto-launch)  	    (write-inst-goto-launch 	(cdr i)))
	((eq? (car i) 'launch-label)  	    (write-inst-launch-label 	(cdr i)))
	((eq? (car i) 'external)  	    (write-inst-external 	(cdr i)))
	((eq? (car i) 'read-scalar)         (write-inst-read-scalar         (cdr i) in-display?))
	((eq? (car i) 'read-array-element)  (write-inst-read-array-element  (cdr i) in-display?))
	((eq? (car i) 'array-empty)         (write-inst-array-empty         (cdr i) in-display?))
	((eq? (car i) 'array-rowcount)      (write-inst-array-rowcount      (cdr i) in-display?))
	((eq? (car i) 'lookup-array-index)  (write-inst-lookup-array-index  (cdr i) in-display?))
	((eq? (car i) 'stack-maxindex)      (write-inst-stack-maxindex      (cdr i) in-display?))

	(else   (if in-display?

		    ;; Instructions that can be used only in display context:

		    (cond ((eq? (car i) 'html)                 (write-inst-html          (cdr i)))
			  ((eq? (car i) 'html-newline)         (write-inst-html-newline  (cdr i)))
			  ((eq? (car i) 'html-field)           (write-inst-html-field    (cdr i)))
			  ((eq? (car i) 'html-textarea)        (write-inst-html-textarea (cdr i)))
			  ((eq? (car i) 'html-button)          (write-inst-html-button   (cdr i)))
			  ((eq? (car i) 'html-table-button)    (write-inst-html-table-button (cdr i)))
			  ((eq? (car i) 'html-launcher-button) (write-inst-html-launcher-button (cdr i))))

		    ;; Instructions that cannot be used in display context:

		    (cond ((eq? (car i) 'authenticate)        (write-inst-authenticate (cdr i) package-name schema-id))
			  ((eq? (car i) 'check-schema)        (write-inst-check-schema        (cdr i)))
			  ((eq? (car i) 'display)             (write-inst-display             (cdr i)))
			  ((eq? (car i) 'write-scalar)        (write-inst-write-scalar        (cdr i)))
			  ((eq? (car i) 'write-array-element) (write-inst-write-array-element (cdr i)))
			  ((eq? (car i) 'generate-rowid)      (write-inst-generate-rowid      (cdr i)))
			  ((eq? (car i) 'delete-var)          (write-inst-delete-var          (cdr i)))
			  ((eq? (car i) 'delete-array-row)    (write-inst-delete-array-row    (cdr i)))
			  ((eq? (car i) 'stack-truncate)      (write-inst-stack-truncate      (cdr i)))
			  ((eq? (car i) 'gosub)               (write-inst-gosub               (cdr i)))
			  ((eq? (car i) 'return)              (write-inst-return              (cdr i))))))))

(define (write-inst-label i)
  (newline)
  (display "   case \"")
  (display i)
  (display "\":")
  (newline)
  (newline))

(define (write-inst-bogus args)
  (display "    # Not implemented yet or bogus input - unable to compile ")
  (display args)
  (newline))

(define (write-inst-comment args)
  (let ((comment-string (car args)))
    (newline)
    (display "    # ")
    (display comment-string)
    (newline)
    (newline)))

(define (write-inst-include args)
  (let ((include-item (car args)))
    (newline)
    (display "    # Placeholder for \"")
    (display include-item)
    (display "\"")
    (newline)
    (newline)))

(define (write-inst-declare-temp args)
  (let ((tempname (car args)))
    (display "    # Declaration of variable $")
    (display tempname)
    (display " omitted (not required in PHP)")
    (newline)))

(define (write-inst-assign args)
  (let ((lhs (car args))
	(rhs (cadr args)))
    (display "    ")
    (write-arg-general lhs)
    (display " = ")
    (write-arg-general rhs)
    (display ";")
    (newline)))

(define (write-inst-append args)
  (let ((lhs (car args))
	(rhs (cadr args)))
    (display "    ")
    (write-arg-general lhs)
    (display " .= ")
    (write-arg-general rhs)
    (display ";")
    (newline)))

(define (write-inst-increment args)
  (let ((lhs (car args))
	(rhs (cadr args)))
    (display "    ")
    (write-arg-general lhs)
    (display " += ")
    (write-arg-general rhs)
    (display ";")
    (newline)))

(define (write-inst-decrement args)
  (let ((lhs (car args))
	(rhs (cadr args)))
    (display "    ")
    (write-arg-general lhs)
    (display " -= ")
    (write-arg-general rhs)
    (display ";")
    (newline)))

(define (write-inst-multiply args)
  (let ((lhs (car args))
	(rhs (cadr args)))
    (display "    ")
    (write-arg-general lhs)
    (display " *= ")
    (write-arg-general rhs)
    (display ";")
    (newline)))

(define (write-inst-divide args)
  (let ((lhs (car args))
	(rhs (cadr args)))
    (display "    ")
    (write-arg-general lhs)
    (display " /= ")
    (write-arg-general rhs)
    (display ";")
    (newline)))

(define (write-inst-substring args)
  (let ((lhs (car args))
	(string (cadr args))
	(offset (caddr args)))
    (display "    ")
    (write-arg-general lhs)
    (display " = substr (")
    (write-arg-general string)
    (display ", ")
    (write-arg-general offset)
    (display ");")
    (newline)))

(define (write-inst-number-equal args)
  (let ((lhs (car args))
	(value1 (cadr args))
	(value2 (caddr args)))
    (display "    ")
    (write-arg-general lhs)
    (display " = (")
    (write-arg-general value1)
    (display " == ")
    (write-arg-general value2)
    (display ");")
    (newline)))

(define (write-inst-number-less args)
  (let ((lhs (car args))
	(value1 (cadr args))
	(value2 (caddr args)))
    (display "    ")
    (write-arg-general lhs)
    (display " = (")
    (write-arg-general value1)
    (display " < ")
    (write-arg-general value2)
    (display ");")
    (newline)))

(define (write-inst-number-greater args)
  (let ((lhs (car args))
	(value1 (cadr args))
	(value2 (caddr args)))
    (display "    ")
    (write-arg-general lhs)
    (display " = (")
    (write-arg-general value1)
    (display " > ")
    (write-arg-general value2)
    (display ");")
    (newline)))

(define (write-inst-string-equal args)
  (let ((lhs (car args))
	(value1 (cadr args))
	(value2 (caddr args)))
    (display "    ")
    (write-arg-general lhs)
    (display " = (")
    (write-arg-general value1)
    (display " === ")
    (write-arg-general value2)
    (display ");")
    (newline)))

(define (write-inst-string-match args)
  (let ((lhs (car args))
	(value (cadr args))
	(pattern (caddr args)))
    (display "    ")
    (write-arg-general lhs)
    (display " = ereg (\"^\" . ")
    (write-arg-general pattern)
    (display ", ")
    (write-arg-general value)
    (display ");")
    (newline)))

(define (write-inst-and args)
  (let ((lhs (car args))
	(value1 (cadr args))
	(value2 (caddr args)))
    (display "    ")
    (write-arg-general lhs)
    (display " = (")
    (write-arg-general value1)
    (display " && ")
    (write-arg-general value2)
    (display ");")
    (newline)))

(define (write-inst-or args)
  (let ((lhs (car args))
	(value1 (cadr args))
	(value2 (caddr args)))
    (display "    ")
    (write-arg-general lhs)
    (display " = (")
    (write-arg-general value1)
    (display " || ")
    (write-arg-general value2)
    (display ");")
    (newline)))

(define (write-inst-not args)
  (let ((lhs (car args))
	(value1 (cadr args)))
    (display "    ")
    (write-arg-general lhs)
    (display " = (! (")
    (write-arg-general value1)
    (display "));")
    (newline)))

(define (write-inst-sqrt args)
  (let ((lhs (car args))
	(value1 (cadr args)))
    (display "    ")
    (write-arg-general lhs)
    (display " = sqrt (")
    (write-arg-general value1)
    (display ");")
    (newline)))

(define (write-inst-goto args)
  (let ((label (car args)))
    (display "    $target = \"")
    (display label)
    (display "\";")
    (newline)
    (display "    break;")
    (newline)
    (newline)))

(define (write-inst-goto-if-true args)
  (let ((label (car args))
	(test (cadr args)))
    (display "    if (")
    (write-arg-general test)
    (display ")")
    (newline)
    (display "    {")
    (newline)
    (display "      $target = \"")
    (display label)
    (display "\";")
    (newline)
    (display "      break;")
    (newline)
    (display "    }")
    (newline)
    (newline)))

(define (write-inst-goto-launch args)
  (let ((mission-id (car args)))
    (display "    $target = \"mission_\" . ")
    (write-arg-general mission-id)
    (display " . \"_launch_selection\";")
    (newline)
    (display "    break;")
    (newline)
    (newline)))
    
(define (write-inst-launch-label args)
  (let ((mission-id (car args)))
    (newline)
    (display "   case \"mission_")
    (write-arg-general mission-id)
    (display "_launch_selection\":")
    (newline)
    (newline)))

(define (write-inst-external args)
  (let ((lhs (car args))
	(subroutine-name (cadr args))
	(arguments (cddr args)))
    (newline)
    (display "    # External call to ")
    (display subroutine-name)
    (display " subroutine")
    (newline)
    (display "    ")
    (write-arg-general lhs)
    (display " = ")
    (display subroutine-name)
    (display " ")

    (cond ((= (length arguments) 0)
	   (display "();"))

	  ((= (length arguments) 1)
	   (display "(")
	   (write-arg-general (car arguments))
	   (display ");"))

	  (else 
	   (display "(")
	   (write-arg-general (car arguments))
	   (for-each (lambda (a)
		       (display ", ")
		       (write-arg-general a))
		     (cdr arguments))
	   (display ");")))
    (newline)
    (newline)))

(define (write-inst-read-scalar args in-display?)
  (let ((lhs (car args))
	(stack-frame (cadr args))
	(variable (caddr args)))
    (display "    ")
    (write-arg-general lhs)
    (display " = ")
    (if in-display?
	(display "var_scalar_read ($state, ")
	(display "var_scalar_read ($next_state, "))
    (write-arg-general stack-frame)
    (display ", ")
    (write-arg-general variable)
    (display ");")
    (newline)))

(define (write-inst-read-array-element args in-display?)
  (let ((lhs (car args))
	(stack-frame (cadr args))
	(variable (caddr args))
	(col-count (cadddr args))
	(row (cadr (cdddr args)))
	(col (caddr (cdddr args))))
    (display "    ")
    (write-arg-general lhs)
    (display " = ")
    (if in-display?
	(display "var_array_element_read ($state, ")
	(display "var_array_element_read ($next_state, "))
    (write-arg-general stack-frame)
    (display ", ")
    (write-arg-general variable)
    (display ", ")
    (write-arg-general col-count)
    (display ", ")
    (write-arg-general row)
    (display ", ")
    (write-arg-general col)
    (display ");")
    (newline)))

(define (write-inst-array-empty args in-display?)
  (let ((lhs (car args))
	(stack-frame (cadr args))
	(variable (caddr args)))
    (display "    ")
    (write-arg-general lhs)
    (display " = ")
    (if in-display?
	(display "var_array_empty ($state, ")
	(display "var_array_empty ($next_state, "))
    (write-arg-general stack-frame)
    (display ", ")
    (write-arg-general variable)
    (display ");")
    (newline)))

(define (write-inst-array-rowcount args in-display?)
  (let ((lhs (car args))
	(stack-frame (cadr args))
	(variable (caddr args))
	(col-count (cadddr args)))
    (display "    ")
    (write-arg-general lhs)
    (display " = ")
    (if in-display?
	(display "var_array_rowcount ($state, ")
	(display "var_array_rowcount ($next_state, "))
    (write-arg-general stack-frame)
    (display ", ")
    (write-arg-general variable)
    (display ", ")
    (write-arg-general col-count)
    (display ");")
    (newline)))

(define (write-inst-lookup-array-index args in-display?)
  (let ((lhs (car args))
	(target (cadr args))
	(stack-frame (caddr args))
	(variable (cadddr args))
	(col-count (cadr (cdddr args))))
    (newline)
    (display "    ")
    (write-arg-general lhs)
    (display " = -1;")
    (newline)
    (display "    for ($i = 0; $i < ")
    (if in-display?
      (display "var_array_rowcount ($state, ")
      (display "var_array_rowcount ($next_state, "))
    (write-arg-general stack-frame)
    (display ", ")
    (write-arg-general variable)
    (display ", ")
    (write-arg-general col-count)
    (display "); $i++)")
    (newline)
    (display "    {")
    (newline)
    (display "      if (")
    (if in-display?
	(display "var_array_element_read ($state, ")
	(display "var_array_element_read ($next_state, "))
    (write-arg-general stack-frame)
    (display ", ")
    (write-arg-general variable)
    (display ", ")
    (write-arg-general col-count)
    (display ", $i, 0) === ")
    (write-arg-general target)
    (display ")")
    (newline)
    (display "      {")
    (newline)
    (display "        ")
    (write-arg-general lhs)
    (display " = $i;")
    (newline)
    (display "      }")
    (newline)
    (display "    }")
    (newline)
    (newline)))

(define (write-inst-stack-maxindex args in-display?)
  (let ((lhs (car args)))
    (display "    ")
    (write-arg-general lhs)
    (display " = ")
    (if in-display?
	(display "var_stack_maxindex ($state)")
	(display "var_stack_maxindex ($next_state)"))
    (display ";")
    (newline)))

(define (write-inst-html args)
  (let ((displaystyle (car args))
	(val (cadr args)))
    (display "    ")
    (display (cond ((null? displaystyle) "$main_html")
		   ((eq? (car displaystyle) 'header)  "$header_html")
		   ((eq? (car displaystyle) 'left)    "$left_html")
		   ((eq? (car displaystyle) 'main)    "$main_html")
		   ((eq? (car displaystyle) 'right)   "$right_html")
		   ((eq? (car displaystyle) 'footer)  "$footer_html")
		   (else "$main_html")))
    (display " .= ")
    (write-arg-general val)
    (display ";")
    (newline)))

(define (write-inst-html-newline args)
  (let ((displaystyle (car args)))
    (display "    ")
    (display (cond ((null? displaystyle) "$main_html")
		   ((eq? (car displaystyle) 'header)  "$header_html")
		   ((eq? (car displaystyle) 'left)    "$left_html")
		   ((eq? (car displaystyle) 'main)    "$main_html")
		   ((eq? (car displaystyle) 'right)   "$right_html")
		   ((eq? (car displaystyle) 'footer)  "$footer_html")
		   (else "$main_html")))
    (display " .= \"\\n\";")
    (newline)))

(define (write-inst-html-field args)
  (let ((displaystyle (car args))
	(name (cadr args))
	(val (caddr args)))
    (display "    ")
    (display (cond ((null? displaystyle) "$main_html")
		   ((eq? (car displaystyle) 'header)  "$header_html")
		   ((eq? (car displaystyle) 'left)    "$left_html")
		   ((eq? (car displaystyle) 'main)    "$main_html")
		   ((eq? (car displaystyle) 'right)   "$right_html")
		   ((eq? (car displaystyle) 'footer)  "$footer_html")
		   (else "$main_html")))
    (display " .= \"<input type=\\\"text\\\" name=\\\"")
    (display name)
    (display "\\\" value=\\\"\" . ")
    (write-arg-general val)
    (display " . \"\\\" />\";")
    (newline)))

(define (write-inst-html-textarea args)
  (let ((displaystyle (car args))
	(name (cadr args))
	(val (caddr args)))
    (display "    ")
    (display (cond ((null? displaystyle) "$main_html")
		   ((eq? (car displaystyle) 'header)  "$header_html")
		   ((eq? (car displaystyle) 'left)    "$left_html")
		   ((eq? (car displaystyle) 'main)    "$main_html")
		   ((eq? (car displaystyle) 'right)   "$right_html")
		   ((eq? (car displaystyle) 'footer)  "$footer_html")
		   (else "$main_html")))
    (display " .= \"<textarea name=\\\"")
    (display name)
    (display "\\\" rows=\\\"10\\\" cols=\\\"60\\\">\" . ")
    (write-arg-general val)
    (display " . \"</textarea>\";")
    (newline)))

(define (write-inst-html-button args)
  (let ((displaystyle (car args))
	(val (cadr args))
	(text (caddr args)))
    (display "    ")
    (display (cond ((null? displaystyle) "$main_html")
		   ((eq? (car displaystyle) 'header)  "$header_html")
		   ((eq? (car displaystyle) 'left)    "$left_html")
		   ((eq? (car displaystyle) 'main)    "$main_html")
		   ((eq? (car displaystyle) 'right)   "$right_html")
		   ((eq? (car displaystyle) 'footer)  "$footer_html")
		   (else "$main_html")))
    (display " .= \"<input class=\\\"controlButton\\\" type=\\\"submit\\\" name=\\\"action\\\" value=\\\"\" . ")
    (write-arg-general text)
    (display " . \"\\\" onClick=\\\"form.control.value = '\" . ")
    (write-arg-general val)
    (display " . \"'\\\" />\";")
    (newline)))

(define (write-inst-html-table-button args)
  (let ((displaystyle (car args))
	(val (cadr args))
	(text (caddr args)))
    (display "    ")
    (display (cond ((null? displaystyle) "$main_html")
		   ((eq? (car displaystyle) 'header)  "$header_html")
		   ((eq? (car displaystyle) 'left)    "$left_html")
		   ((eq? (car displaystyle) 'main)    "$main_html")
		   ((eq? (car displaystyle) 'right)   "$right_html")
		   ((eq? (car displaystyle) 'footer)  "$footer_html")
		   (else "$main_html")))
    (display " .= \"<input class=\\\"tableButton\\\" type=\\\"submit\\\" name=\\\"action\\\" value=\\\"\" . ")
    (write-arg-general text)
    (display " . \"\\\" onClick=\\\"form.control.value = '\" . ")
    (write-arg-general val)
    (display " . \"'\\\" />\";")
    (newline)))

(define (write-inst-html-launcher-button args)
  (let ((displaystyle (car args))
	(val (cadr args))
	(text (caddr args)))
    (display "    ")
    (display (cond ((null? displaystyle) "$main_html")
		   ((eq? (car displaystyle) 'header)  "$header_html")
		   ((eq? (car displaystyle) 'left)    "$left_html")
		   ((eq? (car displaystyle) 'main)    "$main_html")
		   ((eq? (car displaystyle) 'right)   "$right_html")
		   ((eq? (car displaystyle) 'footer)  "$footer_html")
		   (else "$main_html")))
    (display " .= \"<input class=\\\"launcherButton\\\" type=\\\"submit\\\" name=\\\"action\\\" value=\\\"\" . ")
    (write-arg-general text)
    (display " . \"\\\" onClick=\\\"form.control.value = '\" . ")
    (write-arg-general val)
    (display " . \"'\\\" />\";")
    (newline)))

(define (write-inst-authenticate args package-name schema-id)
  (newline)
  (display "    # Authenticate based on submitted login/password")
  (newline)
  (newline)
  (display "    $login = $parameters ['login'];")
  (newline)
  (display "    $password = $parameters ['password'];")
  (newline)
  (newline)
  (display "    list ($name, $uid, $current_cid) = db_read_user ($dbh, \"")
  (display package-name)
  (display "\", $login, $password);")
  (newline)
  (display "    if ($uid != 0)")
  (newline)
  (display "    {")
  (newline)
  (display "      $restored_state = db_read_continuation ($dbh, $current_cid, 0);")
  (newline)
  (newline)
  (display "      if (! $restored_state)")
  (newline)
  (display "      {")
  (newline)
  (display "        # If we got to this point, authentication succeeded but no valid continuation was")
  (newline)
  (display "        # found. Go to beginning of program.")
  (newline)
  (newline)
  (display "        $next_state ['parent_uid'] = $uid;")
  (newline)
  (display "        $next_state ['resumed_session'] = 'no';")
  (newline)
  (display "        $target = 'main';")
  (newline)
  (display "        break;")
  (newline)
  (display "      }")
  (newline)
  (newline)
  (display "      if ($restored_state ['schema_id'] !== \"")
  (display schema-id)
  (display "\")")
  (newline)
  (display "      {")
  (newline)
  (display "        # Authentication succeeded and continuation was found, but schema is not current.")
  (newline)
  (display "        # Go to beginning of program.")
  (newline)
  (newline)
  (display "        $next_state ['parent_uid'] = $uid;")
  (newline)
  (display "        $next_state ['resumed_session'] = 'schema_reset';")
  (newline)
  (display "        $target = 'main';")
  (newline)
  (display "        break;")
  (newline)
  (display "      }")
  (newline)
  (newline)
  (display "      # Authentication succeeded, valid continuation was found, and schema is")
  (newline)
  (display "      # current. Proceed from the display specified in the continuation record.")
  (newline)
  (newline)
  (display "      $next_state = $restored_state;")
  (newline)
  (display "      $next_state ['resumed_session'] = 'yes';")
  (newline)
  (newline)
  (display "      $restored_display_id = $next_state ['display_state'];")
  (newline)
  (display "      display_state_update ($next_state, $restored_display_id);")
  (newline)
  (display "      return $next_state;")
  (newline)
  (display "    }")
  (newline)
  (newline)
  (display "    # If we got to this point, authentication failed.")
  (newline)
  (display "    # Clear parent uid and advance to \"bad login\" display.")
  (newline)
  (newline)
  (display "    $next_state ['parent_uid'] = 0;")
  (newline)
  (newline)
  (display "    # End of authentication code.")
  (newline)
  (newline))

(define (write-inst-check-schema args)
  (newline)
  (display "    # Check that the current schema has been written to the database; if not, write it now:")
  (newline)
  (display "    check_update_schema ($dbh);")
  (newline)
  (newline))

(define (write-inst-display args)
  (let ((display-id (car args))
	(display-name (cadr args)))
    (newline)
    (display "    # Set current display ID to ")
    (display display-id)
    (display " (")
    (display display-name)
    (display "),")
    (newline)
    (display "    # then return to caller to save state and generate HTML.")
    (newline)
    (newline)
    (display "   case \"top_of_display_")
    (display display-id)
    (display "\":")
    (newline)
    (display "    display_state_update ($next_state, ")
    (display display-id)
    (display ");")
    (newline)
    (display "    return $next_state;")
    (newline)
    (newline)
    (display "   case \"resume_from_display_")
    (display display-id)
    (display "\":")
    (newline)
    (newline)))

(define (write-inst-write-scalar args)
  (let ((val (car args))
	(stack-frame (cadr args))
	(variable (caddr args)))
    (display "    ")
    (display "var_scalar_write ($next_state, ")
    (write-arg-general stack-frame)
    (display ", ")
    (write-arg-general variable)
    (display ", ")
    (write-arg-general val)
    (display ");")
    (newline)))

(define (write-inst-write-array-element args)
  (let ((val (car args))
	(stack-frame (cadr args))
	(variable (caddr args))
	(col-count (cadddr args))
	(row (cadr (cdddr args)))
	(col (caddr (cdddr args))))
    (display "    ")
    (display "var_array_element_write ($next_state, ")
    (write-arg-general stack-frame)
    (display ", ")
    (write-arg-general variable)
    (display ", ")
    (write-arg-general col-count)
    (display ", ")
    (write-arg-general row)
    (display ", ")
    (write-arg-general col)
    (display ", ")
    (write-arg-general val)
    (display ");")
    (newline)))

(define (write-inst-generate-rowid args)
  (let ((lhs (car args)))
    (display "    ")
    (write-arg-general lhs)
    (display " = gen_rowid ();")
    (newline)))

(define (write-inst-delete-var args)
  (let ((stack-frame (car args))
	(variable (cadr args)))
    (display "    var_delete ($next_state, ")
    (write-arg-general stack-frame)
    (display ", ")
    (write-arg-general variable)
    (display ");")
    (newline)))

(define (write-inst-delete-array-row args)
  (let ((stack-frame (car args))
	(variable (cadr args))
	(col-count (caddr args))
	(row (cadddr args)))
    (display "    var_array_deleterow ($next_state, ")
    (write-arg-general stack-frame)
    (display ", ")
    (write-arg-general variable)
    (display ", ")
    (write-arg-general col-count)
    (display ", ")
    (write-arg-general row)
    (display ");")
    (newline)))

(define (write-inst-stack-truncate args)
  (let ((stack-frame (car args)))
    (display "    var_stack_truncate ($next_state, ")
    (write-arg-general stack-frame)
    (display ");")
    (newline)))

(define (write-inst-gosub args)
  (let ((lhs (car args))
	(subroutine-name (cadr args))
	(subroutine-id (caddr args))
	(subroutine-label (cadddr args))
	(return-location (cadr (cdddr args)))
	(arguments (cddr (cdddr args))))
    (newline)
    (display "    # Subroutine call site ")
    (display return-location)
    (display ". Calling subroutine \"")
    (display subroutine-name)
    (display "\", id ")
    (display subroutine-id)
    (display ".")
    (newline)
    (display "    # Push activation record onto stack, then goto ")
    (display subroutine-label)
    (display ".")
    (newline)
    (newline)
    (display "    subroutine_push ($next_state, ")
    (display subroutine-id)
    (display ", ")
    (display return-location)
    (display ", array (")

    (if (> (length arguments) 0)
	(begin
	  (write-arg-general (car arguments))
	  (for-each (lambda (arg)
		      (display ", ")
		      (write-arg-general arg))
		    (cdr arguments))))

    (display "));")
    (newline)
    (display "    $target = \"")
    (display subroutine-label)
    (display "\";")
    (newline)
    (display "    break;")
    (newline)
    (display "   case \"resume_from_subroutine_call_")
    (display return-location)
    (display "\":")
    (newline)
    (newline)
    (display "    # Read return value from slot 2 of the top activation record:")
    (newline)
    (newline)
    (display "    ")
    (write-arg-general lhs)
    (display " = var_scalar_read ($next_state, var_stack_maxindex ($next_state), 2);");
    (newline)
    (display "    subroutine_pop ($next_state);")
    (newline)
    (newline)))

(define (write-inst-return args)
  (newline)
  (display "    # Return from subroutine:")
  (newline)
  (newline)
  (display "    $target = \"resume_from_subroutine_call_\" . var_scalar_read ($next_state, var_stack_maxindex ($next_state), 1);")
  (newline)
  (display "    break;")
  (newline)
  (newline))

(define (write-arg-general arg)
  (cond ((eq? (car arg) 'const)    (write-arg-const (cdr arg)))
	((eq? (car arg) 'temp)     (write-arg-temp  (cdr arg)))
	((eq? (car arg) 'field)    (write-arg-field (cdr arg)))))

(define (write-arg-const args)
  (let ((value (car args)))
    (write value)))

(define (write-arg-temp args)
  (let ((name (car args)))
    (display "$")
    (display name)))

(define (write-arg-field args)
  (let ((name (car args)))
    (display "$parameters ['")
    (display name)
    (display "']")))

;;;----------------------------------------------------------------------
;;; Emit the PHP code that updates the database schema information
;;; if necessary.
;;;----------------------------------------------------------------------

(define (write-schema-update-code schema-id schema-entries)
  (display "function check_update_schema ($dbh)")
  (newline)
  (display "{")
  (newline)
  (display "    # Lock schema table to avoid the possibility of accidentally")
  (newline)
  (display "    # writing out the schema more than once.")
  (newline)
  (newline)
  (display "    mysql_query (\"LOCK TABLES schema WRITE\", $dbh);")
  (newline)
  (newline)
  (display "    $sth = mysql_query (\"SELECT COUNT(*) FROM schema WHERE schema_id = '")
  (display schema-id)
  (display "'\", $dbh);")
  (newline)
  (newline)
  (display "    $row = mysql_fetch_row ($sth);")
  (newline)
  (display "    $rowcount = $row [0];")
  (newline)
  (newline)
  (display "    mysql_free_result ($sth);")
  (newline)
  (newline)
  (display "    if ($rowcount == 0)")
  (newline)
  (display "    {")
  (newline)
  (display "        # If the subroutine_id is 0, the entry is for a static (global) variable.")
  (newline)
  (display "        # If the subroutine_id is 1, the entry is for a display (mapping display name to id, etc.)")
  (newline)
  (display "        # If the subroutine_id is 2, the entry is for a subroutine (mapping subroutine name to id, etc.)")
  (newline)
  (display "        # If the subroutine_id is >=64, the entry is for a subroutine argument or local variable.")
  (newline)
  (display "        # In this latter case, the first argument/local is always at index 3, the second at index 4, etc.")
  (newline)
  (newline)

  (for-each (lambda (entry)
	      (display "        mysql_query (\"INSERT INTO schema (schema_id, variable, subroutine_id, variable_name, type_name, type_index, field_name, field_type, field_label) VALUES ('")
	      (display schema-id)
	      (display "', ")
	      (display (get-value 'variable entry))
	      (display ", ")
	      (display (get-value 'subroutine-id entry))
	      (display ", '")
	      (display (get-value 'variable-name entry))
	      (display "', '")
	      (display (get-value 'type-name entry))
	      (display "', ")
	      (display (get-value 'type-index entry))
	      (display ", ")
	      (let ((field-name (get-value 'field-name entry)))
		(if field-name
		    (begin
		      (display "'")
		      (display field-name)
		      (display "'"))
		    (display "NULL")))
	      (display ", ")
	      (let ((field-type (get-value 'field-type entry)))
		(if field-type
		    (begin
		      (display "'")
		      (display field-type)
		      (display "'"))
		    (display "NULL")))
	      (display ", ")
	      (let ((field-label (get-value 'field-label entry)))
		(if field-label
		    (begin
		      (display "'")
		      (display field-label)
		      (display "'"))
		    (display "NULL")))
	      (display ")\", $dbh);")
	      (newline))
	    schema-entries)

  (display "    }")
  (newline)
  (newline)
  (display "    mysql_query (\"UNLOCK TABLES\", $dbh);")
  (newline)
  (display "}")
  (newline))

;;;----------------------------------------------------------------------
;;; End of procedure definitions. Run the main program.
;;;----------------------------------------------------------------------

(main)
